#!/usr/bin/Rscript

# This file provides a nice template for LDA analysis that can be easily extended and modified.
# It provides basic functionality so that is easy to run simple LDA analysises.

# The code works by:
# 1) loading all data from a particular directory,
# 2) preproceses it,
# 3) builds topic model for provided list of K values,
# 4) evaluates all topic models based on harmonic mean measure to pick
# the best value of K,
# 5) prints for each topic a selected number of words
# 6) prints a list of associated documents for each topic
# 7) optionally for each topic, prints n-th line of each
# document. This can be for example a documents year, title or author
# if all documnets follow the same format. (e.g., to have author name
# of the first line)
#
# Authors: Vitomir Kovanovic,    Srećko Joksimovi?
#          v.kovanovic@ed.ac.uk  sjoksimo@sfu.ca
# Addtional comments: Go ota
# The original programs are used in LAK 2013 workshop
# Modified:
# Modified for Japanese with using RMeCab by Go Ota  2015/06/22

Sys.setlocale("LC_COLLATE", "C")

suppressWarnings(suppressMessages(library(tm)))
suppressWarnings(suppressMessages(library(lda)))
suppressWarnings(suppressMessages(library(slam)))
suppressWarnings(suppressMessages(library(topicmodels)))
suppressWarnings(suppressMessages(library(stringr)))
suppressWarnings(suppressMessages(library(xtable)))
suppressWarnings(suppressMessages(library(Rmpfr)))

#
# this function writes provided str to the screen
#
log.message <- function(str) {
    cat(date(), "\t", str, "\n")
}

#
# generic processing function that takes as an input anothre function
# (fn parameter) and a directory and calls fn for each file in a
# directory. This can be any type of processing, such as loading file
# content, or loading particular line of the file, or anything
# completely different.
#
process.all.files <- function(fn, load.dir) {
    results <- list()
    file.names <- list()

    for(input.file in list.files(path=load.dir)) {
        results <- append(results, fn(load.dir, input.file))
        file.names <- append(file.names, input.file)
    }

    list(results=unlist(results),
         file.names=unlist(file.names))
}

#
# processing function that is used together with process.all.files.
# this processing function receives a directory path and filename and
# it reads the complete file content from a file with a given name in
# a load.dir directory
#
process.single.file.contents <- function(load.dir, f) {
    d.lines <- readLines(paste0(load.dir, f))
	print(f)
 #   d <- tolower(paste(d.lines, collapse=" "))
     d <- paste(d.lines, collapse=" ")
}

#
# this function loads all files in a load.dir directory.
#
load.all.file.contents <- function(load.dir) {
    log.message(paste("Loading file contents from directory", load.dir))

    loading.res <- process.all.files(process.single.file.contents, load.dir)
    list(all.file.contents=loading.res$results, all.file.names=loading.res$file.names)
}

# this function is used to build a document-term matrix (DTM) from the
# text files that are found in the load.dir. All other arguments
# specify particular filtering steps:
#
# tf.threshold: number between 0.0 and 1.0 that specifies the maximum
# sparcity of the term row. For example 0.95 means to filter out all
# rows that do not appear in at least 5% of the documents. we
# typically do this to make DTM smaller and faster and also to remove
# words that would not help in interpretation
#
# tf.idf.threshold: number between 0.0 and 1.0 that is used to filter
# out words that have too low "importance" given by TF-IDF threshold.
#
# tf.idf.function: value "mean" or "median". Defines which function is
# used for TF.IDF filtering.
#
# For example tf.idf.threshold of 0.9 and
# tf.idf.function 'mean', it means filter out all terms that do not
# have TF-IDF at leat 0.9*mean value. This way we can specify to
# remove only the worst of the worst (e.g, 0.5*median).
#
# stemming: T or F, indicating whether word stemming should be
# done. Stemming is a proces of finding non-changing word root. For
# example for word produce, stem is produc- as there are variations
# produced, production etc, but all of them have 'produc' at the
# start. This is generally done in order to have less terms in the
# matrix and to match similar terms that are used to describe similar
# things.
#
# stopwords.removal: T or F, indicating whether removal of very
# frequently used words (e.g., a, an, the, I, can) should be
# done. Those words typically do not contribute to any meaning and are
# just plain linguistic 'fillers'.
#
# minWordLength: integer number (1, 2, 3...) indicating the minimal
# length of the word. This can be used to remove very short words that
# do not bring any value to topic modeling (e.g., OK, do, be)
#
# removeNumbers: T or F indicating whether numbers should be removed
# from the input text. This is typically done as they are not very
# meaningful as topical terms.
#
# removePunctuation: T or F indicating whether non alphanumerical
# characters should be removed from the input.
#
# minDocFreq: integer number (1, 2, 3..) that indicates the minimal
# number of documents in which words should apear. For example,
# documents typically have personal names and lastnames, and if they
# do not appear in other documents, they should be removed as they are
# not meaningful for describing topics, if they only appear in one
# document.
#
build.dtm <- function(load.dir,
                      tf.threshold=F,
                      tf.idf.threshold=F,
                      tf.idf.function = "median",
                      stemming=F,
                      stopwords.removal=F,
                      minWordLength=1,
                      removeNumbers=F,
                      removePunctuation=F,
                      minDocFreq=1) {

    data <- load.all.file.contents(load.dir)$all.file.contents

    dtm <- DocumentTermMatrix(Corpus(VectorSource(data)),
                              control = list(
                                  stemming = stemming,
                                  stopwords = stopwords.removal,
                                  minWordLength = minWordLength,
                                  removeNumbers = removeNumbers,
                                  removePunctuation = removePunctuation,
                                  minDocFreq = minDocFreq))

    log.message(paste("Original number of terms:", dim(dtm)[2]))
    print(dtm)

    if(tf.threshold) {
        log.message(paste("Running TF filtering with TF sparcity threshold of", tf.threshold))
        dtm <- removeSparseTerms(dtm, tf.threshold)

        log.message(paste("After TF filtering:", dim(dtm)[2]))
        print(dtm)
    }

    if(tf.idf.threshold) {
        log.message(paste("Running TF-IDf filtering with TF-IDF threshold of", tf.idf.threshold, "and with function", tf.idf.function))

        term.tfidf <- tapply(dtm$v/row_sums(dtm)[dtm$i], dtm$j, mean) * log2(nDocs(dtm)/col_sums(dtm > 0))

        tf.idf.run <- T
        tf.idf.base.val <- 1

        if(tf.idf.function == "median")
            tf.idf.base.val <- median(term.tfidf)

        else if(tf.idf.function == "mean")
            tf.idf.base.val <- mean(term.tfidf)

        else {
            log.message(paste0("WARNING: wrong function for TF-IDF filering specified: ", tf.idf.function, ". available functions are mean and median. Skipping TF-IDF filtering"))
            tf.idf.run <- F
        }
        if (tf.idf.run) {
            computed.value <- tf.idf.threshold * tf.idf.base.val
            log.message(paste("TF-IDF computed threshold", computed.value))

            dtm <- dtm[,term.tfidf >= computed.value]
            dtm <- dtm[row_sums(dtm) > 0,]

            log.message(paste("After TF-IDF filtering:", dim(dtm)[2]))
        }
    }

    print(dtm)

    return(dtm)
}

#
# Actual function that runs LDA algorithm for a specified
# document-term matrix, K, and alpha values.
#
run.lda <- function(k, dtm, alpha) {

    # Several parameters related to Gibbs sampling are specified here,
    # refer to topicmodels manual for their meaning.
    burnin <- 0
    iter <- 2000
    keep <- 50
    seed <- 1
    alpha <- alpha

#   LDA(dtm, k = k, method = "Gibbs", control = llist(burnin, iter, keep, seed))
#	modified Go.Ota 2015/06/22
#Change No.1 	If you want to change value of alpha, please change the next line directory;
	LDA(dtm, k = k, method = "Gibbs", control = list(burnin = 0, iter = 2000, keep = 50, seed = 1, alpha = 1.0))
}

#
# Function that builds topic models from document-term matrix for all
# values of K provided in the vector all.ks and with a given dirichlet
# prior alpha.
#
build.topic.models <- function(dtm, all.ks, alpha=1.0, fileExist) {
    full.model.file.name <- paste0("models/", project.id, ".RData")
 #   if (file.exists(full.model.file.name)) {
	if (fileExist) {
		log.message(paste0(project.id, " exists, loading..." ))
        load(full.model.file.name)

    } else {
        log.message(paste0('"', project.id, '"', " does not exists, building topic models..." ))

        models <- list()
        for(k in all.ks) {
            log.message(paste("K=", k, "started"))
            models[as.character(k)] <- run.lda(k, dtm, alpha)
        }
        save(models, file = full.model.file.name)
    }

    models
}

#
# Function that calculates harmonic mean from the list of
# logLikelihoods for a set of topic mddels for different values of K
#
harmonicMean <- function(logLikelihoods) {
    precision <- 2000L
    llMed <- median(logLikelihoods)
    as.double(llMed - log(mean(exp(-mpfr(logLikelihoods, prec = precision) + llMed))))
}

#
# Function that evaluates list of topic models to find the one with
# the best topic model based on the harmonic mean metric.
#
evaluate.topic.models <- function(topic.models) {
    all.ks <- as.integer(names(all.topic.models))
    log.message("Evaluating all topic models")

    model.harmonic.means <- sapply(unlist(lapply(topic.models,logLik)), harmonicMean)
# Modified  Go. Ota 2015/06/22 Under 3
	log.message("model.harmonic.means")
	# // log.message(str(model.harmonic.means))
	write.csv(model.harmonic.means, file = "Harmonic_means_all.csv")
	
    return(list(k=all.ks[which.max(model.harmonic.means)], scores=model.harmonic.means, all.ks=all.ks))
}

#
# for given topic.model and K, return for each document the most
# likely topic. If a document is equally likely to belong to several
# topics, then all of them are returned
#
get.topic.assignments <- function(best, selected.k) {
    gammaDF <- as.data.frame(best@gamma)
    names(gammaDF) <- 1:length(names(gammaDF))

    as.data.frame(cbind(document = row.names(gammaDF), topic = apply(gammaDF,1,function(x) names(gammaDF)[which(x==max(x))])))
}

#
# for a given topic assigment (ta), best topic model (best) print
# n.terms of each topic together with other statistics for each topic.
#
print.topics.and.term <- function(ta, best, n.terms){
    topic.freqs <- sort(table(unlist(ta$topic)), decreasing=T) # count number of documents for each topic

    total.count <- sum(topic.freqs)
    selected.k <- length(topic.freqs)

    cum.counts <- NULL

    for (i in seq(1, length(topic.freqs)))
        cum.counts <- c(cum.counts, sum(topic.freqs[1:i])/total.count)

    terms <- terms(best, n.terms)

    topic.names.old <- names(topic.freqs)[1:selected.k]

    imp.terms <- data.frame(terms[,as.numeric(topic.names.old)])
    imp.terms <- data.frame(do.call(paste, c(data.frame(t(imp.terms)), sep=", ")))
    imp.terms[,2] <- topic.names.old
    imp.terms[,3] <- round(cum.counts, 2)
    imp.terms[,4] <- as.character(topic.freqs[1:selected.k])

    imp.terms <- imp.terms[,c(2, 4, 3, 1)]
    colnames(imp.terms) <- c("topic", "articles", "articles cummulative", paste(n.terms, "terms"))

    cat("\n", n.terms, "most important terms for each topic.\n")
    print(imp.terms)

    print(xtable(imp.terms),
          file=paste0("out/", project.id, "_topic_terms.html"),
          type="html",
          include.rownames=F,
          sanitize.text.function=function(x){x})
}

#
# function that makes a line graph that can be used to pick optimal
# number of topics. It receives the result of all topic.model
# evaluations (res)
#
plot.ks <- function(res) {
    all.ks <- res$all.ks

    scores <- res$scores
    best.k <- res$k

    pdf(paste0("out/", project.id, "_topic_evaluation.pdf"), width=3, height=2.5)
    par(list(oma=c(0, 0, 0, 0), mai=c(0.50, 0.30, 0.10, 0.05), cex=0.6))

    plot(all.ks, scores, type = "l", xlab = "Number of topics", axes=F, ylab="")
    axis(1, at=c(2,25,50,75,100,125,150,175,200), col.axis="black", tick=T, las=1)
    axis(2, at=range(scores), labels=F,           col.axis="black", las=1)
    title(ylab = "Log-likelihood", mgp=c(1,0,0))

    if(is.null(best.k)==F) {
        abline(v=best.k, col="red", lty=3)

        y.text.pos <- range(scores)[1] + (0.8*(range(scores)[2]-range(scores)[1]))

        text(x=best.k,
             y=y.text.pos,
             labels=as.character(best.k))
    }

    invisible(dev.off())
}

#
# for a given topic assigments (ta), topic.id, list of filenames
#
print.nth.lines.for.topic <- function(ta, topic.id, all.file.names, headlines) {

    all.file.names <-unlist(all.file.names)

    target.docs <- unlist(lapply(ta$topic,
                                 function(topics) {
                                     return(as.character(topic.id) %in% topics)
                                 }))

    target.doc.names <- all.file.names[target.docs]

    cat("\n\tThere are", length(target.doc.names), "files primarily about topic", topic.id, "\n\n")
    cat(headlines[str_detect(headlines, paste0("^", paste(target.doc.names, collapse="|")))], sep="\n")
}

#
# for a given topic assigments (ta) and topic.id, print list of files
# associated with this topic.
#
print.topic.filenames <- function(ta, topic.id, all.file.names) {

    all.file.names <-unlist(all.file.names)
	
	# debug 2015.05.11
#	cat ("all.file.names******")
#	cat (all.file.names)
	
    target.docs <- unlist(lapply(ta$topic,
                                 function(topics) {
                                     return(as.character(topic.id) %in% topics)
                                 }))

    target.doc.names <- all.file.names[target.docs]

    cat("\n\tThere are", length(target.doc.names), "files primarily about topic", topic.id, "\n\n")
    cat(target.doc.names, sep="\n")
}

#
# function that reads n-th line of every file that is found in the load.dir
#
load.nth.row <- function(load.dir, n) {
    log.message(paste("Loading", n, "row from each file"))

    # a function that reads n-th line of file f from a directory
    # load.dir. The function is nested so that the value of argument
    # n from the outher function is visible.
    process.single.file.headline <- function(load.dir, f) {
        paste0("\t", f, ": ", readLines(paste0(load.dir, f))[n])
    }

    loading.res <- process.all.files(process.single.file.headline, load.dir)
    list(all.file.headlines=loading.res$results, all.file.names=loading.res$file.names)
}

#
# function that reads n-th line of every file that is found in the load.dir
#
load.filenames <- function(load.dir, n) {
    log.message(paste("Loading file names from", load.dir))

    # a function that does nothing, as we're interested only in filenames
    process.nothing <- function(load.dir, f) {}

    loading.res <- process.all.files(process.nothing, load.dir)
    loading.res$file.names
}

#
# function that prints for each topic n-th line of topic's associated
# files. This can be used for example, to print 1st line of each file
# which can be a document's title (if the data is structured like
# that).

# Arguments:
# ta: topic assignemts for all documents in the corpus
# load.dir: directory from where to load n-th line
# n: the file line that is to be loaded
#
print.nth.line <- function(ta, load.dir, n=1) {
    topic.freqs <- sort(table(unlist(ta$topic)), decreasing=T)

    selected.k <- length(topic.freqs)


    all.files.nth.line.res <- load.nth.row(load.dir, n)
    all.files.nth.lines     <- all.files.nth.line.res$all.file.headlines
    all.files.names         <- all.files.nth.line.res$all.file.names

    cat("\nTopic headlines:\n")
    for(topic.id in names(topic.freqs)) {
        cat("\nTopic", topic.id, "\n")
        print.nth.lines.for.topic(ta, topic.id, all.files.names, all.files.nth.lines)
    }
}

print.topics.filenames <- function(ta, load.dir) {
    topic.freqs <- sort(table(unlist(ta$topic)), decreasing=T)

    selected.k <- length(topic.freqs)

    all.files.names <- load.filenames(load.dir)

    cat("\nTopic headlines:\n")
    for(topic.id in names(topic.freqs)) {
        cat("\nTopic", topic.id, "\n")
        print.topic.filenames(ta, topic.id, all.files.names)
    }
}


#
# from the list of all.topic.models, pick the one that is best
# according to harmonic mean metric. Also make the evaluation plot tht
# shows harmonic means metric for all models.
#
get.best.model <- function(all.topic.models) {
    evaluation.results <- evaluate.topic.models(all.topic.models) # calculate all harmonic mean scores and find best K

    if(length(evaluation.results$all.ks)>1)
        log.message("Generating topic evaluation plot")
        plot.ks(evaluation.results)

    all.topic.models[[as.character(evaluation.results$k)]]
}

# add Go.Ota
# write topic list to files
write.topics <- function(topics, filename){
	topics.no <- length(topics)
	t_topics <- topics
	log.message("start of check")
	for( i in 1:topics.no){
		templ = length(t_topics[[i]])
		if(templ == 2){
			t_topics[[i]] <- c( t_topics[[i]][1], t_topics[[i]][2],0)
		}
		else if(templ == 1){
			t_topics[[i]] <- c( t_topics[[i]][1], 0, 0)
		}
		else if(templ == 0){
			t_topics[[i]] <- c( 0, 0, 0)
		}
		# log.message(t_topics[[i]])
	}
	log.message("end of check")
	write.csv(t_topics, file = filename)

}

